c*******************************************************************
c*                         COMBINE                                 *
c*                                                                 *
c*  Short description:                                             *
c*    Combines results from MC calculations. Intended to be used   *
c*    for simulations done with the same algorithm but different   *
c*    RNG seeds (e.g. for parallel calculations).                  *
c*                                                                 *
c*    COMBINE reads its configuration data from the keyboard and   *
c*    writes to the screen. Therefore, it is usually executed by   *
c*    redirecting these devices as in                              *
c*     $ combine.x < combine.in > combine.out                      *
c*                                                                 *
c*    Lines starting with '#' in an output file are processed      *
c*    after removing the comment mark. This provides support       *
c*    for files adapted to gnuplot scripting which also need to be *
c*    read by COMBINE.                                             *
c*                                                                 *
c* Last update:                                                    *
c*    2005-02-28  A. Badal                                         *
c*      The treatment of (x,y,Dy) data sets has been improved.     *
c*    2003-12-22  Created (result)                                 *
c*                                                                 *
c* Comments:                                                       *
c*   - Adapt the code to more clones or more data by changing the  *
c*     parameters MAXCLN and MAXDAT throughout this file. Notice   *
c*     that this implies an amount of memory approx equal to       *
c*     (MAXDAT * 16 + 528) * MAXCLN * 1e-6 MB of RAM               *
c*     (currently some 34 MB).                                     *
c*                                                                 *
c* Josep Sempau, email: josep.sempau@upc.es                        *
c* Universitat Politecnica de Catalunya, Barcelona, Spain          *
c*******************************************************************
      program main
      implicit none
      integer maxcln,maxdat
      parameter (maxcln=1024,maxdat=2048)
      logical done(maxcln)
      character*256 nick(maxcln),outfn(maxcln),reskw,qoikw,nkw,tkw
      integer nclon,ndata,nskip,ncol,xcol,qoicol,unccol
      real*8 qoi(maxcln,maxdat),uncert(maxcln,maxdat),x(maxdat)
      real*8 nmin,tmin,uncmax,n(maxcln),t(maxcln)

c       Read input:
      call rdin(maxcln,nick,outfn,nclon,reskw,qoikw,nkw,tkw,
     &          nskip,ncol,xcol,qoicol,unccol,nmin,tmin,uncmax)

c       Read output files:
      call rdout(nclon,outfn,reskw,qoikw,nkw,tkw,nskip,ncol,
     &           xcol,qoicol,unccol,done,ndata,x,qoi,uncert,n,t)

c       Write results:
      call wrres(nclon,nick,done,ndata,x,qoi,uncert,n,t,
     &           nmin,tmin,uncmax)
      stop
      end


      subroutine rdin(maxcln,nick,outfn,nclon,reskw,qoikw,nkw,tkw,
     &                nskip,ncol,xcol,qoicol,unccol,nmin,tmin,uncmax)
c*******************************************************************
c*    Reads input data.                                            *
c*                                                                 *
c*    Input:                                                       *
c*      maxcln -> max no. of clons                                 *
c*    Output:                                                      *
c*      nick -> array of clon nicknames                            *
c*      outfn -> array of output file names                        *
c*      nclon -> no. of clons                                      *
c*      reskw -> results keyword                                   *
c*      qoikw -> QOI keyword                                       *
c*      nkw -> N keyword                                           *
c*      tkw -> t keyword                                           *
c*      nskip -> no. of lines to skip after QOI pattern is found   *
c*      ncol -> no. of columns to read in report                   *
c*      xcol -> no. of the column where the abscissa for QOI is    *
c*      qoicol -> no. of the column where QOI is                   *
c*      unccol -> no. of the column where uncertainty is           *
c*      nmin -> min N                                              *
c*      tmin -> min t                                              *
c*      uncmax -> max uncertainty (%)                              *
c*    Comments:                                                    *
c*      -> for default values, see below.                          *
c*******************************************************************
      implicit none
      integer maxcln,nclon,nskip,ncol,xcol,qoicol,unccol
      character*(*) nick(maxcln),outfn(maxcln),reskw,qoikw,nkw,tkw
      real*8 nmin,tmin,uncmax

      character*256 buffer,suffix
      integer error,strlen,i,str1st

c     *** Defaults:
      nclon = 0
      reskw = ' '
      qoikw = ' '
      nkw = ' '
      tkw = ' '
      nskip = 0
      ncol = 2
      xcol = 0
      qoicol = 1
      unccol = 2
      nmin = 0.0d0
      tmin = 0.0d0
      uncmax = 1.0d30

 10   continue
        call getlin(buffer,error)
        if (error.ne.0) then
          write(*,*) 'rdin:ERROR: likely, missing EndOfInput.'
          stop
        endif

        if (index(buffer,'[ClonTable]').ne.0) then
c         *** Read clon nicknames:
          call clnick(maxcln,nick,nclon)

        else if (index(buffer,'[OutputSuffix]').ne.0) then
          read(*,'(a)') suffix
          if (strlen(suffix).eq.0) then
            write(*,*) 'rdin:WARNING: void suffix.'
          else
            suffix = suffix(str1st(suffix):)
          endif

        else if (index(buffer,'[ResultsKeyword]').ne.0) then
          read(*,'(a)') reskw
          if (strlen(reskw).eq.0) then
            write(*,*) 'rdin:ERROR: void results keyword.'
            stop
          else
            reskw = reskw(str1st(reskw):)
          endif

        else if (index(buffer,'[QOI-keyword]').ne.0) then
          read(*,'(a)') qoikw
          if (strlen(qoikw).eq.0) then
            write(*,*) 'rdin:ERROR: void QOI keyword.'
            stop
          else
            qoikw = qoikw(str1st(qoikw):)
          endif

        else if (index(buffer,'[skip]').ne.0) then
          read(*,*) nskip

        else if (index(buffer,'[columns]').ne.0) then
          read(*,*) xcol,qoicol,unccol
          ncol = max(xcol,qoicol,unccol)
        else if (index(buffer,'[N-keyword]').ne.0) then
          read(*,'(a)') nkw
          if (strlen(nkw).eq.0) then
            write(*,*) 'rdin:ERROR: void N keyword.'
            stop
          else
            nkw = nkw(str1st(nkw):)
          endif

        else if (index(buffer,'[t-keyword]').ne.0) then
          read(*,'(a)') tkw
          if (strlen(tkw).eq.0) then
            write(*,*) 'rdin:ERROR: void t keyword.'
            stop
          else
            tkw = tkw(str1st(tkw):)
          endif

        else if (index(buffer,'[N-min]').ne.0) then
          read(*,*) nmin

        else if (index(buffer,'[t-min]').ne.0) then
          read(*,*) tmin

        else if (index(buffer,'[UncertMax]').ne.0) then
          read(*,*) uncmax

        else if (index(buffer,'[EndOfInput]').ne.0) then
          if (strlen(reskw).eq.0) then
            write(*,*) 'rdin:ERROR: results keyword not provided.'
            stop
          endif
          if (strlen(qoikw).eq.0) then
            write(*,*) 'rdin:ERROR: QOI keyword not provided.'
            stop
          endif
          if (strlen(nkw).eq.0) then
            write(*,*) 'rdin:ERROR: N keyword not provided.'
            stop
          endif
c         *** Construct output file names:
          do i=1,nclon
            if (strlen(suffix).ne.0) then
              outfn(i) = nick(i)(:strlen(nick(i)))//'.'//suffix
            else
              outfn(i)= nick(i)
            endif
          enddo
          return

        else
          write(*,*) 'rdin:ERROR: incomprehensible input:'
          write(*,*) buffer
          stop
        endif

      goto 10
      end


      subroutine rdout(nclon,outfn,reskw,qoikw,nkw,tkw,nskip,ncol,
     &                 xcol,qoicol,unccol,done,ndata,x,qoi,uncert,
     &                 n,t)
c*******************************************************************
c*    Reads output files.                                          *
c*                                                                 *
c*    Input:                                                       *
c*      nclon -> no. of clons (i.e. array dimensions)              *
c*      outfn -> array of output file names                        *
c*      reskw -> results keyword                                   *
c*      qoikw -> QOI keyword                                       *
c*      nkw -> N keyword                                           *
c*      tkw -> t keyword                                           *
c*      nskip -> no. of lines to skip after QOI keyword is found   *
c*      ncol -> no. of columns to read in report                   *
c*      xcol -> no. of the column where the abscissa for the QOI is*
c*      qoicol -> no. of the column where QOI is                   *
c*      unccol -> no. of the column where uncertainty is           *
c*    Output:                                                      *
c*      done -> array of data status; .true. if all data read      *
c*      ndata -> no. of data in output files (taken from 1st)      *
c*      x -> array of abscissa values                              *
c*      qoi -> array of QOI values                                 *
c*      uncert -> array of uncertainties                           *
c*      n -> array of no. of simulated histories                   *
c*      t -> array of simulation times                             *
c*******************************************************************
      implicit none
      integer maxcln,maxdat
      parameter (maxcln=1024,maxdat=2048)
      integer nclon,ndata,nskip,ncol,xcol,qoicol,unccol
      logical done(nclon)
      character*(*) outfn(nclon),reskw,qoikw,nkw,tkw
      real*8 qoi(maxcln,maxdat),uncert(maxcln,maxdat),x(maxdat)
      real*8 n(nclon),t(nclon)

      logical doneq,donen,donet, xerror
      character*256 buffer
      integer error,strlen,i,funit,finduf,j,k
      integer maxcol
      parameter (maxcol=10)
      real*8 dum(maxdat,maxcol), x_temp

      funit = finduf()
      ndata = 0

      clone: do i=1,nclon
c       *** Defaults:
        doneq = .false.
        donen = .false.
        donet = .false.
        done(i) = .false.
        t(i) = 0.0d0
        xerror = .false.

        write(*,*) 'rdout: opening file: ',outfn(i)(:strlen(outfn(i)))
        open(unit=funit,file=outfn(i),status='old',iostat=error)
        if (error.ne.0) then
          write(*,*) 'rdout:WARNING: unable to open file.'
          cycle clone
        endif

c     *** Locate results section in output file:
        do
          call fgetln(funit,buffer,error)
          if (error.ne.0) goto 10
          if (index(buffer,reskw(:strlen(reskw))).ne.0) exit
        enddo

c       *** Read simulation data:
        keyword: do
          call fgetln(funit,buffer,error)
          if (error.ne.0) exit keyword

          if (.not.doneq.and.
     &        index(buffer,qoikw(:strlen(qoikw))).ne.0) then
c           *** QOI section; skip dummy lines first:
            do j=1,nskip
              read(funit,'(a)') buffer
            enddo
            row: do j=1,maxdat+1
              if (j.eq.maxdat+1) then
                write(*,*) ' ERROR: too many data; enlarge MAXDAT'
                stop
              endif
              call fgetln(funit,buffer,error)
              read(buffer,*,iostat=error) (dum(j,k),k=1,ncol)
              if (error.eq.0) then
                qoi(i,j) = dum(j,qoicol)
                uncert(i,j) = dum(j,unccol)
                if (xcol.eq.0) exit row  ! No abscissa-> one single row
                x_temp = dum(j,xcol)
                if (ndata.eq.0) then
c               ** Store the abscissa values from the first non-empty clone output file
                  x(j) = x_temp
                else
c               ** Compare the abscissa values to those from the first non-empty clone
c                  output file, taking into account the different rounding error with
c                  various compilers (maximum relative difference of 0.01%).
                  if (abs(x_temp-x(j)).gt.1.0d-4*x(j)) xerror = .true.
                endif
              else
                exit row  ! End of data set
              endif
            enddo row
            if (xcol.ne.0) j = j-1
c           *** Take no. of data from first non-empty clone output file:
            if (ndata.eq.0.and.j.ne.0) then
              ndata = j
              write(*,*)
     &        '  No. of data rows to be read set to: ',j
            else if (j.ne.ndata) then
              write(*,*)
     &        '  WARNING: unable to read QOI; No. data rows found:',j
              exit keyword
            endif

            if (xerror) then
            write(*,*)
     &      '  WARNING: The abscissa values found are not',
     &      ' consistent with the values from the first clone read.'
            write(*,*)
     &      ' Please make sure that the combined files correspond to',
     &      ' the same simulation.'
              exit keyword   ! Do not use the present clone
            endif

            if (j.gt.0) doneq = .true.
          endif

          if (.not.donen.and.
     &        index(buffer,nkw(:strlen(nkw))).ne.0) then
            call fgetln(funit,buffer,error)  ! Supports gnuplot comment mark removal
            read(buffer,*,iostat=error) n(i)
            if (error.ne.0) then
              write(*,*) '  WARNING: unable to read No. histories.'
              exit keyword
            endif
            donen = .true.
          endif

          if (.not.donet.and.strlen(tkw).ne.0.and.
     &        index(buffer,tkw(:strlen(tkw))).ne.0) then
            call fgetln(funit,buffer,error)  ! Supports gnuplot comment mark removal
            read(buffer,*,iostat=error) t(i)
            if (error.ne.0) then
              write(*,*) '  WARNING: unable to read simulation time.'
              t(i) = 0.0d0
              exit keyword
            endif
            donet = .true.
          endif
        enddo keyword

 10     continue
        if (doneq.and.donen) done(i) = .true.
        if (.not.done(i)) write(*,*)
     &  '  WARNING: unable to get essential data from report section'
        close(funit)
      enddo clone
      end


      subroutine wrres(nclon,nick,done,ndata,x,qoi,uncert,n,t,
     &                 nmin,tmin,uncmax)
c*******************************************************************
c*    Writes overall result.                                       *
c*                                                                 *
c*    Input:                                                       *
c*      nclon -> no. of clons (i.e. array dimensions)              *
c*      nick -> array of clon nicknames                            *
c*      done -> array of data status; .true. if all data read      *
c*      ndata -> no. of data in output files                       *
c*      x -> array with the abscissa values for the QOI            *
c*      qoi -> array of QOI values                                 *
c*      uncert -> array of uncertainties                           *
c*      n -> array of no. of simulated histories                   *
c*      t -> array of simulation times                             *
c*      nmin -> min N                                              *
c*      tmin -> min t                                              *
c*      uncmax -> max uncertainty (%)                              *
c*******************************************************************
      implicit none
      integer maxcln,maxdat
      parameter (maxcln=1024,maxdat=2048)
      integer nclon,ndata
      logical done(nclon)
      character*(*) nick(nclon)
      real*8 qoi(maxcln,maxdat),uncert(maxcln,maxdat),x(maxdat)
      real*8 nmin,tmin,uncmax,n(nclon),t(nclon)

      character*256 slow,fast
      integer k,strlen,len1,j
      real*8 qtot(ndata),unctot(ndata),ntot,ttot,speed,allunc,efftot
      real*8 qoibin(ndata),uncbin(ndata),clnspd,minspd,maxspd,maxq

      write(*,*) ' '
      write(*,*) 'wrres: [OVERALL RESULT REPORT] '

c     *** Compute averages:
      minspd = 1.0d30
      slow = 'No clon reported simulation time'
      maxspd = 0.0d0
      fast = 'No clon reported simulation time'
      ntot = 0.0d0
      ttot = 0.0d0
      speed = 0.0d0
      do j=1,ndata
        qtot(j)   = 0.0d0
        unctot(j) = 0.0d0
      enddo

      do k=1,nclon
        len1 = strlen(nick(k))
c       *** Skip if not done yet:
        if (.not.done(k)) then
          write(*,*) '  WARNING: ',nick(k)(:len1),' not done;skipped.'
          goto 10
        endif

c       *** Check thresholds:
        if (n(k).lt.nmin) then
          write(*,*) '  WARNING: no. histories below minimum in ',
     &      nick(k)(:len1),' ; skipped. N: '
          write(*,*) int(n(k))
          goto 10
        endif
        if (t(k).lt.tmin) then
          write(*,*) '  WARNING: running time below minimum in ',
     &      nick(k)(:len1),' ; skipped. t: '
          write(*,'(3x,1pe12.5)') t(k)
          goto 10
        endif
c       *** Find average uncertainty:
        do j=1,ndata
          qoibin(j) = qoi(k,j)
          uncbin(j) = uncert(k,j)
        enddo
        call aveunc(ndata,qoibin,uncbin,maxq,allunc)
        if (allunc.gt.uncmax) then
          write(*,*) '  WARNING: uncertainty above maximum in ',
     &      nick(k)(:len1),' ; skipped. Unc(%): '
          write(*,'(3x,1pe12.5)') allunc
          goto 10
        endif
        if (t(k).gt.0.0d0) then
          clnspd = n(k)/t(k)
          speed = speed+clnspd
          if (clnspd.gt.maxspd) then
            maxspd = clnspd
            fast = nick(k)
          endif
          if (clnspd.lt.minspd) then
            minspd = clnspd
            slow = nick(k)
          endif
        else
          write(*,*) '  WARNING: ',nick(k)(:len1),
     &               ' reported null or void simulation time.'
        endif

        ntot = ntot+n(k)
        ttot = ttot+t(k)
        do j=1,ndata
          qtot(j) = qtot(j)+n(k)*qoi(k,j)
          unctot(j) = unctot(j)+(n(k)*uncert(k,j))**2
        enddo
 10     continue
      enddo

      if (ntot.lt.0.5d0) then
        write(*,*) '  '
        write(*,*) '  ERROR: no clon reported.'
        stop
      endif
      do j=1,ndata
        qtot(j) = qtot(j)/ntot
        unctot(j) = sqrt(unctot(j))/ntot
      enddo

c     *** Write results:
      write(*,*) ' '
      if (ndata.eq.1) then
        write(*,*) '    QOI      +- uncert'
        write(*,'(3x,1pe12.5,2x,1pe8.2)') qtot(1),unctot(1)
      else
        write(*,*) '    abscissa        QOI      +- uncert'
        write(*,'(3x,1pe12.5,2x,1pe12.5,2x,1pe8.2)')
     &       (x(j), qtot(j), unctot(j), j=1,ndata)
      endif
      write(*,*) ' '
      write(*,*) '  Overall performance report.'
      write(*,*) '  No. of histories simulated [N]:'
      if (ntot.lt.2.1d9) then
        write(*,'(3x,i10)') int(ntot+0.1d0)
      else
        write(*,'(3x,1pe21.14)') ntot
      endif
      write(*,*)
     &  '  Total CPU time [t]--taking all clones into account:'
      write(*,'(3x,1pe12.5)') ttot
      write(*,*)
     &  '  Speed (hist/RealTime)--if clones were in single-process:'
      write(*,'(3x,1pe12.5)') speed
      len1 = strlen(fast)
      write(*,'(3x,a,1pe11.5,a,a)')
     &  '  Max speed (',maxspd,') reached in: ',fast(:len1)
      len1 = strlen(slow)
      write(*,'(3x,a,1pe11.5,a,a)')
     &  '  Min speed (',minspd,') reached in: ',slow(:len1)
      write(*,*) '  Uncertainty in % [uncert]:'
      call aveunc(ndata,qtot,unctot,maxq,allunc)
      write(*,'(3x,1pe12.5)') allunc
      write(*,*) '  Intrinsic efficiency [uncert^2*N]^-1 :'
      if (allunc.gt.0.0d0) then
        efftot = 1.0d0/(ntot*allunc**2)
      else
        efftot = 0.0d0
      endif
      write(*,'(3x,1pe12.5)') efftot
      write(*,*)
     &  '  Absolute efficiency [uncert^2*RealTime]^-1 :'
      efftot = efftot*speed
      write(*,'(3x,1pe12.5)') efftot
      write(*,*) ' '
      write(*,*) 'Have a nice day.'
      end


      subroutine aveunc(ndata,qoi,uncert,maxq,allunc)
c*******************************************************************
c*    Computes overall uncertainty for a list of QOIs and uncerts. *
c*                                                                 *
c*    Input:                                                       *
c*      ndata -> array dimensions                                  *
c*      qoi -> array of QOIs                                       *
c*      uncert -> array of uncertainties                           *
c*    Output:                                                      *
c*      maxq -> max abs(QOI) --positive value--                    *
c*      allunc -> overall relative uncertainty, in %               *
c*    Comments:                                                    *
c*      - The relative uncertainty (RU) is in percentage. This RU  *
c*        is the sqrt of the average of the squared RUs computed   *
c*        over all bins for which abs(QOI) exceeds 1/2 of the      *
c*        maximum QOI.                                             *
c*******************************************************************
      implicit none
      integer ndata
      real*8 qoi(ndata),uncert(ndata),maxq,allunc

      integer i,nbin
      real*8 absqoi

      allunc = 0.0d0
      maxq = 0.0d0
      do i=1,ndata
        absqoi = abs(qoi(i))
        if (absqoi.gt.maxq) maxq = absqoi
      enddo
      if (maxq.eq.0.0d0) return

      nbin = 0
      do i=1,ndata
        absqoi = abs(qoi(i))
        if (absqoi.gt.0.5d0*maxq) then
          allunc = allunc+(uncert(i)/absqoi)**2
          nbin = nbin+1
        endif
      enddo
      allunc = 100.0d0*sqrt(allunc/nbin)
      end


      subroutine clnick(maxcln,nick,nclon)
c*******************************************************************
c*    Reads clon nicknames from stdin                              *
c*                                                                 *
c*    Input:                                                       *
c*      maxcln -> max no. of clons                                 *
c*    Output:                                                      *
c*      nick -> array of clon nicknames                            *
c*      nclon -> no. of clons                                      *
c*******************************************************************
      implicit none
      integer maxcln,nclon
      character*(*) nick(maxcln)

      character*256 buffer
      integer k,j,error,pos

      k = 1
 10   continue
        if (k.gt.maxcln) then
          write(*,*) 'clnick:ERROR: too many clons in clon table.'
          stop
        endif
        call getlin(buffer,error)
        if (error.ne.0) then
          write(*,*) 'clnick:ERROR: likely, missing EndOfTable.'
          stop
        endif

c       *** Check EOT condition:
        if (index(buffer,'[EndOfTable]').ne.0) then
          nclon = k-1
          return
        endif

c       *** Parse nickname:
        call capstr(buffer,nick(k),pos)
        if (pos.eq.0) then
          write(*,*) 'clnick:ERROR: unable to read nickname from input:'
          write(*,*) buffer
          stop
        endif

c       *** Check consistency:
        do j=1,k-1
          if (nick(k).eq.nick(j)) then
            write(*,*) 'clnick:ERROR: clons with same nickname:'
            write(*,*) nick(k)
            stop
          endif
        enddo
        k = k+1
      goto 10
      end


      subroutine capstr(string,substr,pos)
c*******************************************************************
c*    Captures the substring of string that starts after the 1st   *
c*    non-blank character and ends when the next blank is found.   *
c*                                                                 *
c*    Input:                                                       *
c*      string -> character string                                 *
c*    Output:                                                      *
c*      substr -> substring                                        *
c*      pos -> last position read from string into substr          *
c*******************************************************************
      implicit none
      character*(*) string,substr
      integer pos

      character nextch
      integer i,j,k

      pos = 0
      substr = ' '
      do j=1,len(string)
        if (string(j:j).ne.' ') goto 10
      enddo
      return
 10   continue

      do k=1,len(string)-j+1
        i = j+k-1
        nextch = string(i:i)
        if (nextch.eq.' ') return
        substr(k:k) = nextch
        pos = i
      enddo
      end


      integer function str1st(string)
c*******************************************************************
c*    Returns position in string of first non-blank character.     *
c*    If string is all blanks, it returns zero.                    *
c*                                                                 *
c*    Input:                                                       *
c*      string -> character string                                 *
c*******************************************************************
      implicit none
      character*(*) string

      do str1st=1,len(string)
        if (string(str1st:str1st).ne.' ') return
      enddo
      str1st = 0
      end


      integer function strlen(string)
c*******************************************************************
c*    Returns length of string excluding trailing blanks           *
c*                                                                 *
c*    Input:                                                       *
c*      string -> character string                                 *
c*******************************************************************
      implicit none
      character*(*) string

      do strlen=len(string),1,-1
        if (string(strlen:strlen).ne.' ') return
      enddo
      strlen = 0
      end


      subroutine getlin(buffer,error)
c*******************************************************************
c*    Reads a non-comment line from stdin                          *
c*                                                                 *
c*    Output:                                                      *
c*      buffer -> line just read                                   *
c*      error -> not 0 if an error has ocurred                     *
c*******************************************************************
      implicit none
      character*(*) buffer
      integer error,strlen

 10   continue
      read(*,'(a)',iostat=error) buffer
      if (error.ne.0) return
      if (buffer(1:1).eq.'#') goto 10
      if (strlen(buffer).eq.0) goto 10
      end


      subroutine fgetln(funit,buffer,error)
c*******************************************************************
c*    Reads a line from a given file unit.                         *
c*    If the line starts with a comment mark, the symbol is        *
c*    discarded and the rest of the the line is processed.         *
c*                                                                 *
c*    Input:                                                       *
c*      funit -> file unit to read from                            *
c*    Output:                                                      *
c*      buffer -> line just read                                   *
c*      error -> not 0 if an error has ocurred                     *
c*******************************************************************
      implicit none
      character*(*) buffer
      integer error,funit,strlen

 10   continue
      read(funit,'(a)',iostat=error) buffer
      if (error.ne.0) return
      ! if (buffer(1:1).eq.'#') goto 10  ! Substituted by next line
      if (buffer(1:1).eq.'#') buffer = buffer(2:)  ! Eliminate only the comment mark
      if (strlen(buffer).eq.0) goto 10
      end


      integer function finduf()
c*******************************************************************
c*    Finds a valid (i.e. unused) unit file.                       *
c*******************************************************************
      implicit none
      logical used
      integer maxuf
      parameter (maxuf=17)

      finduf = 6
 10   finduf = finduf+1
        if (finduf.gt.maxuf) then
          write(*,*) ' '
          write(*,*) 'finduf:ERROR: Unable to find a valid unit.'
          stop
        endif
        inquire(finduf,opened=used)
      if (used) goto 10
      end


c >>>> EOF >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

